import numpy as np

from scipy.ndimage.filters import gaussian_filter, laplace
from scipy.signal import medfilt2d


def enhance_peaks(img, clipping_value, gaussian_sigma=2.5):
    type_in = img.dtype
    filtered_img = medfilt2d(img.astype(np.float64))
    filtered_img = gaussian_filter(filtered_img, gaussian_sigma)
    filtered_img = laplace(filtered_img)
    filtered_img *= -1

    filtered_img -= filtered_img.min()
    filtered_img /= filtered_img.max()

    # To leave room to label in the 3d detection.
    filtered_img *= clipping_value
    return filtered_img.astype(type_in)



import numpy as np
import tensorflow as tf

from random import shuffle
from skimage.io import imread
from scipy.ndimage import zoom
from tensorflow.python.keras.utils.data_utils import Sequence
from imlib.cells.cells import group_cells_by_z
from imlib.general.numerical import is_even

from cellfinder_core.classify.augment import AugmentationParameters, augment

# TODO: rename, as now using dask arrays -
#  actually should combine to one generator


class StackSizeError(Exception):
    pass


class CubeGeneratorFromFile(Sequence):
    """
    Reads cubes (defined as e.g. xml, csv) from raw data to pass to
    keras.Model.fit_generator() or keras.Model.predict_generator()

    If augment=True, each augmentation selected has an "augment_likelihood"
    chance of being applied to each cube
    """

    # TODO: shuffle within (and maybe between) batches
    # TODO: limit workers based on RAM

    def __init__(
        self,
        points,
        signal_array,
        background_array,
        voxel_sizes,
        network_voxel_sizes,
        batch_size=16,
        cube_width=50,
        cube_height=50,
        cube_depth=20,
        channels=2,  # No other option currently
        classes=2,
        extract=False,
        train=False,
        augment=False,
        augment_likelihood=0.1,
        flip_axis=[0, 1, 2],
        rotate_max_axes=[1, 1, 1],  # degrees
        # scale=[0.5, 2],  # min, max
        translate=[0.05, 0.05, 0.05],
        shuffle=False,
        interpolation_order=2,
    ):
        self.points = points
        self.signal_array = signal_array
        self.background_array = background_array
        self.batch_size = batch_size
        self.axis_2_pixel_um = float(voxel_sizes[2])
        self.axis_1_pixel_um = float(voxel_sizes[1])
        self.axis_0_pixel_um = float(voxel_sizes[0])
        self.network_axis_2_pixel_um = float(network_voxel_sizes[2])
        self.network_axis_1_pixel_um = float(network_voxel_sizes[1])
        self.network_axis_0_pixel_um = float(network_voxel_sizes[0])
        self.cube_width = cube_width
        self.cube_height = cube_height
        self.cube_depth = cube_depth
        self.channels = channels
        self.classes = classes

        # saving training data to file
        self.extract = extract

        self.train = train
        self.augment = augment
        self.augment_likelihood = augment_likelihood
        self.flip_axis = flip_axis
        self.rotate_max_axes = rotate_max_axes
        # self.scale = scale
        self.translate = translate
        self.shuffle = shuffle
        self.interpolation_order = interpolation_order

        self.scale_cubes = False

        self.rescaling_factor_axis_2 = 1
        self.rescaling_factor_axis_1 = 1
        self.rescaled_cube_width = self.cube_width
        self.rescaled_cube_height = self.cube_height

        self.__check_image_sizes()
        self.__get_image_size()
        self.__check_z_scaling()
        self.__check_in_plane_scaling()
        self.__remove_outlier_points()
        self.__get_batches()
        if shuffle:
            self.on_epoch_end()

    def __check_image_sizes(self):
        if len(self.signal_array) != len(self.background_array):
            raise ValueError(
                f"Number of signal images ({len(self.signal_array)} does not"
                f"match the number of background images "
                f"({len(self.background_array)}"
            )

    def __get_image_size(self):
        self.image_z_size = len(self.signal_array)
        self.image_height, self.image_width = self.signal_array[0].shape

    def __check_in_plane_scaling(self):
        if self.axis_2_pixel_um != self.network_axis_2_pixel_um:
            self.rescaling_factor_axis_2 = (
                self.network_axis_2_pixel_um / self.axis_2_pixel_um
            )
            self.rescaled_cube_width = (
                self.cube_width * self.rescaling_factor_axis_2
            )
            self.scale_cubes = True
        if self.axis_1_pixel_um != self.network_axis_1_pixel_um:
            self.rescaling_factor_axis_1 = (
                self.network_axis_1_pixel_um / self.axis_1_pixel_um
            )
            self.rescaled_cube_height = (
                self.cube_height * self.rescaling_factor_axis_1
            )
            self.scale_cubes = True

    def __check_z_scaling(self):
        if self.axis_0_pixel_um != self.network_axis_0_pixel_um:
            plane_scaling_factor = (
                self.network_axis_0_pixel_um / self.axis_0_pixel_um
            )
            self.num_planes_needed_for_cube = round(
                self.cube_depth * plane_scaling_factor
            )
        else:
            self.num_planes_needed_for_cube = self.cube_depth

        if self.num_planes_needed_for_cube > self.image_z_size:
            raise StackSizeError(
                "The number of planes provided is not sufficient "
                "for any cubes to be extracted. Please check the "
                "input data"
            )

    def __remove_outlier_points(self):
        """
        Remove points that won't get extracted (i.e too close to the edge)
        """
        self.points = [
            point for point in self.points if self.extractable(point)
        ]

    def extractable(self, point):
        x0, x1, y0, y1, z0, z1 = self.__get_boundaries()
        if (
            point.z < z0
            or point.z > z1
            or point.x < x0
            or point.x > x1
            or point.y < y0
            or point.y > y1
        ):
            return False
        else:
            return True

    def __get_boundaries(self):
        x0 = int(round((self.cube_width / 2) * self.rescaling_factor_axis_2))
        x1 = int(round(self.image_width - x0))

        y0 = int(round((self.cube_height / 2) * self.rescaling_factor_axis_1))
        y1 = int(round(self.image_height - y0))

        z0 = int(round(self.num_planes_needed_for_cube / 2))
        z1 = self.image_z_size - z0
        return x0, x1, y0, y1, z0, z1

    def __get_batches(self):
        self.points_groups = group_cells_by_z(self.points)
        # TODO: add optional shuffling of each group here
        self.batches = []
        for centre_plane in self.points_groups.keys():
            points_per_plane = self.points_groups[centre_plane]
            for i in range(0, len(points_per_plane), self.batch_size):
                self.batches.append(points_per_plane[i : i + self.batch_size])

        self.ordered_points = []
        for batch in self.batches:
            for cell in batch:
                self.ordered_points.append(cell)

    def __len__(self):
        """
        Number of batches
        :return: Number of batches per epoch
        """
        return len(self.batches)

    def __getitem__(self, index):
        """
        Generates a single batch of cubes
        :param index:
        :return:
        """
        cell_batch = self.batches[index]
        signal_stack, background_stack = self.__get_stacks(index)
        images = self.__generate_cubes(
            cell_batch, signal_stack, background_stack
        )

        if self.train:
            batch_labels = [cell.type - 1 for cell in cell_batch]
            batch_labels = tf.keras.utils.to_categorical(
                batch_labels, num_classes=self.classes
            )
            return images, batch_labels
        elif self.extract:
            batch_info = self.__get_batch_dict(cell_batch)
            return images, batch_info
        else:
            return images

    def __get_stacks(self, index):
        centre_plane = self.batches[index][0].z

        min_plane, max_plane = get_cube_depth_min_max(
            centre_plane, self.num_planes_needed_for_cube
        )

        signal_stack = np.array(self.signal_array[min_plane:max_plane])
        background_stack = np.array(self.background_array[min_plane:max_plane])

        return signal_stack, background_stack

    def __generate_cubes(self, cell_batch, signal_stack, background_stack):
        number_images = len(cell_batch)
        images = np.empty(
            (
                (number_images,)
                + (self.cube_height, self.cube_width, self.cube_depth)
                + (self.channels,)
            )
        )

        for idx, cell in enumerate(cell_batch):
            images = self.__populate_array_with_cubes(
                images, idx, cell, signal_stack, background_stack
            )

        return images

    def __populate_array_with_cubes(
        self, images, idx, cell, signal_stack, background_stack
    ):
        if self.augment:
            self.augmentation_parameters = AugmentationParameters(
                self.flip_axis,
                self.translate,
                self.rotate_max_axes,
                self.interpolation_order,
                self.augment_likelihood,
            )
        images[idx, :, :, :, 0] = self.__get_oriented_image(cell, signal_stack)
        images[idx, :, :, :, 1] = self.__get_oriented_image(
            cell, background_stack
        )
        return images

    def __get_oriented_image(self, cell, image_stack):
        x0 = int(round(cell.x - (self.rescaled_cube_width / 2)))
        x1 = int(x0 + self.rescaled_cube_width)
        y0 = int(round(cell.y - (self.rescaled_cube_height / 2)))
        y1 = int(y0 + self.rescaled_cube_height)
        image = image_stack[:, y0:y1, x0:x1]
        image = np.moveaxis(image, 0, 2)

        if self.augment:
            # scale to isotropic, but don't scale back
            image = augment(
                self.augmentation_parameters, image, scale_back=False
            )

        pixel_scalings = [
            self.cube_height / image.shape[0],
            self.cube_width / image.shape[1],
            self.cube_depth / image.shape[2],
        ]

        # TODO: ensure this is always the correct size
        image = zoom(image, pixel_scalings, order=self.interpolation_order)
        return image

    @staticmethod
    def __get_batch_dict(cell_batch):
        return [cell.to_dict() for cell in cell_batch]

    def on_epoch_end(self):
        """
        Shuffle data for each epoch
        :return: Shuffled indexes
        """
        shuffle(self.batches)


class CubeGeneratorFromDisk(Sequence):
    """
    Reads in cubes from a list of paths for keras.Model.fit_generator() or
    keras.Model.predict_generator()

    If augment=True, each augmentation selected has a 50/50 chance of being
    applied to each cube
    """

    def __init__(
        self,
        signal_list,
        background_list,
        labels=None,  # only if training or validating
        batch_size=16,
        shape=(50, 50, 20),
        channels=2,
        classes=2,
        shuffle=False,
        augment=False,
        augment_likelihood=0.1,
        flip_axis=[0, 1, 2],
        rotate_max_axes=[45, 45, 45],  # degrees
        # scale=[0.5, 2],  # min, max
        translate=[0.2, 0.2, 0.2],
        train=False,  # also return labels
        interpolation_order=2,
    ):
        self.im_shape = shape
        self.batch_size = batch_size
        self.labels = labels
        self.signal_list = signal_list
        self.background_list = background_list
        self.channels = channels
        self.classes = classes
        self.augment = augment
        self.augment_likelihood = augment_likelihood
        self.flip_axis = flip_axis
        self.rotate_max_axes = rotate_max_axes
        # self.scale = scale
        self.translate = translate
        self.train = train
        self.interpolation_order = interpolation_order
        self.indexes = np.arange(len(self.signal_list))
        if shuffle:
            self.on_epoch_end()

    # TODO: implement scale and shear

    def on_epoch_end(self):
        """
        Shuffle data for each epoch
        :return: Shuffled indexes
        """
        self.indexes = np.arange(len(self.signal_list))
        np.random.shuffle(self.indexes)

    def __len__(self):
        """
        Number of batches
        :return: Number of batches per epoch
        """
        return int(np.ceil(len(self.signal_list) / self.batch_size))

    def __getitem__(self, index):
        """
        Generates a single batch of cubes
        :param index:
        :return:
        """
        # Generate indexes of the batch
        start_index = index * self.batch_size
        end_index = start_index + self.batch_size
        indexes = self.indexes[start_index:end_index]

        # Get data corresponding to batch
        list_signal_tmp = [self.signal_list[k] for k in indexes]
        list_background_tmp = [self.background_list[k] for k in indexes]

        images = self.__generate_cubes(list_signal_tmp, list_background_tmp)

        if self.train:
            batch_labels = [self.labels[k] for k in indexes]
            batch_labels = tf.keras.utils.to_categorical(
                batch_labels, num_classes=self.classes
            )
            return images, batch_labels
        else:
            return images

    def __generate_cubes(self, list_signal_tmp, list_background_tmp):
        number_images = len(list_signal_tmp)
        images = np.empty(
            ((number_images,) + self.im_shape + (self.channels,))
        )

        for idx, signal_im in enumerate(list_signal_tmp):
            background_im = list_background_tmp[idx]
            images = self.__populate_array_with_cubes(
                images, idx, signal_im, background_im
            )

        return images.astype(np.float16)

    def __populate_array_with_cubes(
        self, images, idx, signal_im, background_im
    ):
        if self.augment:
            self.augmentation_parameters = AugmentationParameters(
                self.flip_axis,
                self.translate,
                self.rotate_max_axes,
                self.interpolation_order,
                self.augment_likelihood,
            )
        images[idx, :, :, :, 0] = self.__get_oriented_image(signal_im)
        images[idx, :, :, :, 1] = self.__get_oriented_image(background_im)

        return images

    def __get_oriented_image(self, image_path):
        # if paths are pathlib objs, skimage only reads one plane
        image = np.moveaxis(imread(image_path), 0, 2)
        if self.augment:
            image = augment(self.augmentation_parameters, image)
        return image


def get_cube_depth_min_max(centre_plane, num_planes_needed_for_cube):
    half_cube_depth = num_planes_needed_for_cube // 2
    min_plane = centre_plane - half_cube_depth

    if is_even(num_planes_needed_for_cube):
        # WARNING: not centered because even
        max_plane = centre_plane + half_cube_depth
    else:
        # centered
        max_plane = centre_plane + half_cube_depth + 1

    return min_plane, max_plane



import tempfile

from pathlib import Path
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

from cellfinder_core.download import models
from cellfinder_core.download.download import amend_cfg


home = Path.home()
DEFAULT_DOWNLOAD_DIRECTORY = home / ".cellfinder"
temp_dir = tempfile.TemporaryDirectory()
temp_dir_path = Path(temp_dir.name)


def download_directory_parser(parser):
    parser.add_argument(
        "--install-path",
        dest="install_path",
        type=Path,
        default=DEFAULT_DOWNLOAD_DIRECTORY,
        help="The path to install files to.",
    )
    parser.add_argument(
        "--download-path",
        dest="download_path",
        type=Path,
        default=temp_dir_path,
        help="The path to download files into.",
    )
    parser.add_argument(
        "--no-amend-config",
        dest="no_amend_config",
        action="store_true",
        help="Don't amend the config file",
    )
    return parser


def model_parser(parser):
    parser.add_argument(
        "--no-models",
        dest="no_models",
        action="store_true",
        help="Don't download the model",
    )
    parser.add_argument(
        "--model",
        dest="model",
        type=str,
        default="resnet50_tv",
        help="The model to use",
    )
    return parser


def download_parser():
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser = model_parser(parser)
    parser = download_directory_parser(parser)
    return parser


def main():
    args = download_parser().parse_args()
    if not args.no_models:
        model_path = models.main(args.model, args.install_path)

    if not args.no_amend_config:
        amend_cfg(new_model_path=model_path)


if __name__ == "__main__":
    main()



import natsort

from os import listdir
from os.path import isfile, join

from imlib.cells.cells import Cell, UntypedCell


class TiffList(object):
    """This class represents list of tiff files. These tiff files are the
     output from the cell extractor plugin
    and are used as training and classification data.
    """

    def __init__(self, ch1_list, channels, label=None):
        """A list of tiff files output by the cell extractor plugin to be
        used in machine learning.
        Expects file names to end with "ch[ch].tif", where [ch] is the
        non-zero-padded channel index.
        Given a list of tiff files for the first channel, it will find the
        corresponding files for the
        channels passed in the [channels] parameter.
        :param ch1_list: List of the tiff files of the first channel.
        :param channels: List of channels to use.
        :param label: Label of the directory (e.g. 2 for cell, 1 for no cell).
        Can be ignored on classification runs.
        """

        self.ch1_list = natsort.natsorted(ch1_list)
        self.label = label
        self.channels = channels

    def make_tifffile_list(self):
        """

        :return: Returns the relevant tiff files as a list of TiffFile objects.
        """
        files = [
            f
            for f in self.ch1_list
            if f.lower().endswith("ch" + str(self.channels[0]) + ".tif")
        ]

        tiff_files = [
            TiffFile(tiffFile, self.channels, self.label) for tiffFile in files
        ]
        return tiff_files


class TiffDir(TiffList):
    """A simplified version of TiffList that uses all tiff files without
    any filtering.
    """

    def __init__(self, tiff_dir, channels, label=None):
        super(TiffDir, self).__init__(
            [
                join(tiff_dir, f)
                for f in listdir(tiff_dir)
                if f.lower().endswith("ch" + str(channels[0]) + ".tif")
            ],
            channels,
            label,
        )


class TiffFile(object):
    """This class represents a multichannel tiff file, with one individual
    file per channel.
    """

    def __init__(self, path, channels, label=None):
        self.path = path
        self.channels = channels
        self.label = label

    def files_exist(self):
        return all([isfile(tif) for tif in self.img_files])

    def as_cell(self, force_typed=True):
        if force_typed:
            return (
                Cell(self.path, -1)
                if self.label is None
                else Cell(self.path, self.label)
            )
        else:
            return (
                UntypedCell(self.path)
                if self.label is None
                else Cell(self.path, self.label)
            )

    @property
    def img_files(self):
        return [self.path[:-5] + str(ch) + ".tif" for ch in self.channels]



# cython: language_level=3


cimport libc.math as cmath
cimport cython

import numpy as np
from cellfinder_core.detect.filters.typedefs cimport uint, ushort

# only for __init__
from cellfinder_core.tools.array_operations import bin_mean_3d
from cellfinder_core.tools.geometry import make_sphere

DEBUG = False



cdef class BallFilter:

    cdef:
        uint THRESHOLD_VALUE, SOMA_CENTRE_VALUE
        uint ball_xy_size, tile_step_width, tile_step_height
        int __current_z
        double overlap_fraction, overlap_threshold

        # Numpy arrays
        double[:,:,:] kernel
        ushort[:,:,:] volume
        unsigned char[:,:,:] good_tiles_mask


    def __init__(self, layer_width, layer_height, ball_xy_size, ball_z_size, overlap_fraction=0.8,
                 tile_step_width=None, tile_step_height=None, threshold_value=None, soma_centre_value=None):
        self.ball_xy_size = ball_xy_size
        self.overlap_fraction = overlap_fraction
        self.tile_step_width = tile_step_width
        self.tile_step_height = tile_step_height

        self.THRESHOLD_VALUE = threshold_value
        self.SOMA_CENTRE_VALUE = soma_centre_value

        # temporary kernel of scaling_factor*ball_x_y size to be then scaled to final ball size
        scaling_factor = 2
        x_upscale_factor = y_upscale_factor = z_upscale_factor = 7  # WARNING: needs to be integer
        temp_kernel_shape = [x_upscale_factor * ball_xy_size, y_upscale_factor * ball_xy_size, z_upscale_factor * ball_z_size]
        tmp_ball_centre_position = [cmath.floor(d / 2) for d in temp_kernel_shape]  # z_centre is xy_centre before resize
        tmp_ball_radius = temp_kernel_shape[0] / 2.0
        tmp_kernel = make_sphere(temp_kernel_shape, tmp_ball_radius, tmp_ball_centre_position)
        tmp_kernel = tmp_kernel.astype(np.float64)
        self.kernel = bin_mean_3d(tmp_kernel, x_upscale_factor, y_upscale_factor, z_upscale_factor)

        assert self.kernel.shape[2] == ball_z_size, 'Kernel z dimension should be {}, got {}'\
            .format(ball_z_size, self.kernel.shape[2])

        self.overlap_threshold = self.overlap_fraction * np.array(self.kernel, dtype=np.float64).sum()

        self.volume = np.empty((layer_width, layer_height, self.kernel.shape[2]), dtype=np.uint16)

        self.good_tiles_mask = np.empty((int(cmath.ceil(layer_width / tile_step_width)),  # TODO: lazy initialisation
                                         int(cmath.ceil(layer_height / tile_step_height)),
                                         self.kernel.shape[2]), dtype=np.uint8)
        self.__current_z = -1

    @property
    def ready(self):
        return self.__current_z == self.volume.shape[2] - 1

    cpdef append(self, ushort[:,:] layer, unsigned char[:,:] mask):
        if DEBUG:
            assert [e for e in layer.shape[:2]] == [e for e in self.volume.shape[:2]],\
                'layer shape mismatch, expected "{}", got "{}"'\
                    .format([e for e in self.volume.shape[:2]], [e for e in layer.shape[:2]])
            assert [e for e in mask.shape[:2]] == [e for e in self.good_tiles_mask.shape[2]], \
                'mask shape mismatch, expected"{}", got {}"'\
                    .format([e for e in self.good_tiles_mask.shape[:2]], [e for e in mask.shape[:2]])
        if not self.ready:
            self.__current_z += 1
        else:
            self.volume = np.roll(self.volume, self.volume.shape[2] - 1, axis=2)  # WARNING: not in place
            self.good_tiles_mask = np.roll(self.good_tiles_mask, self.good_tiles_mask.shape[2] - 1, axis=2)
        self.volume[:, :, self.__current_z] = layer[:,:]
        self.good_tiles_mask[:, :, self.__current_z] = mask[:,:]

    cdef get_middle_plane_idx(self):
        cdef uint middle
        middle = <uint> cmath.floor(self.volume.shape[2] / 2)
        return middle

    def get_middle_plane(self):
        cdef uint middle_plane_idx = self.get_middle_plane_idx()
        return np.array(self.volume[:, :, middle_plane_idx], dtype=np.uint16)

    @cython.initializedcheck(False)
    @cython.cdivision(True)
    @cython.boundscheck(False)
    cpdef walk(self):  # Highly optimised because most time critical
        cdef uint stack_middle = <uint> cmath.floor(self.volume.shape[2] / 2)
        cdef uint ball_centre_x, ball_centre_y
        cdef uint ball_radius = self.ball_xy_size // 2
        cdef ushort[:,:,:] cube

        cdef uint max_width, max_height
        tile_mask_covered_img_width = self.good_tiles_mask.shape[0] * self.tile_step_width
        tile_mask_covered_img_height = self.good_tiles_mask.shape[1] * self.tile_step_height
        max_width = tile_mask_covered_img_width - self.ball_xy_size   # whole ball size because the cube is extracted with x + whole ball width
        max_height = tile_mask_covered_img_height - self.ball_xy_size   # whole ball size because the cube is extracted with y + whole ball height
        cdef uint x, y
        for y in range(max_height):
            for x in range(max_width):
                ball_centre_x = x + ball_radius
                ball_centre_y = y + ball_radius
                if self.__is_tile_to_check(ball_centre_x, ball_centre_y):
                    cube = self.volume[x:x + self.kernel.shape[0], y:y + self.kernel.shape[1], :]
                    if self.__cube_overlaps(cube):
                        self.volume[ball_centre_x, ball_centre_y, stack_middle] = self.SOMA_CENTRE_VALUE

    @cython.initializedcheck(False)
    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef __cube_overlaps(self, ushort[:,:,:] cube):  # Highly optimised because most time critical
        """

        :param np.ndarray cube: The thresholded array to check for ball fit. values at CellDetector.THRESHOLD_VALUE are threshold
        :return: True if the overlap exceeds self.overlap_fraction
        """
        if DEBUG:
            assert cube.max() <= 1
            assert cube.shape == self.kernel.shape

        cdef double current_overlap_value = 0

        cdef uint x, y, z
        for z in range(cube.shape[2]):  # TODO: OPTIMISE: step from middle to outer boundaries to check more data first
            if z == cmath.floor(self.volume.shape[2] / 2) + 1 and current_overlap_value < self.overlap_threshold * 0.4:  # FIXME: do not hard code value
                return False  # DEBUG: optimisation attempt
            for y in range(cube.shape[1]):
                for x in range(cube.shape[0]):
                    if cube[x, y, z] >= self.THRESHOLD_VALUE:  # includes self.SOMA_CENTRE_VALUE
                        current_overlap_value += self.kernel[x, y, z]
        return <bint> (current_overlap_value > self.overlap_threshold)

    @cython.initializedcheck(False)
    @cython.cdivision(True)
    @cython.boundscheck(False)
    cdef __is_tile_to_check(self, uint x, uint y):  # Highly optimised because most time critical
        cdef uint x_in_mask, y_in_mask, middle_plane_idx
        x_in_mask = x // self.tile_step_width  # TEST: test bounds (-1 range)
        y_in_mask = y // self.tile_step_height  # TEST: test bounds (-1 range)
        middle_plane_idx = <uint> cmath.floor(self.volume.shape[2] / 2)
        return <bint> self.good_tiles_mask[x_in_mask, y_in_mask, middle_plane_idx]



from datetime import datetime
import multiprocessing
from multiprocessing import Queue as MultiprocessingQueue
from multiprocessing import Lock

import numpy as np

from imlib.general.system import (
    get_num_processes,
)

from cellfinder_core.detect.filters.plane.multiprocessing import (
    MpTileProcessor,
)
from cellfinder_core.detect.filters.setup_filters import setup_tile_filtering
from cellfinder_core.detect.filters.volume.multiprocessing import (
    Mp3DFilter,
)


def calculate_parameters_in_pixels(
    voxel_sizes,
    soma_diameter_um,
    max_cluster_size_um3,
    ball_xy_size_um,
    ball_z_size_um,
):
    """
    Convert the command-line arguments from real (um) units to pixels
    """

    mean_in_plane_pixel_size = 0.5 * (
        float(voxel_sizes[2]) + float(voxel_sizes[1])
    )
    voxel_volume = (
        float(voxel_sizes[2]) * float(voxel_sizes[1]) * float(voxel_sizes[0])
    )
    soma_diameter = int(round(soma_diameter_um / mean_in_plane_pixel_size))
    max_cluster_size = int(round(max_cluster_size_um3 / voxel_volume))
    ball_xy_size = int(round(ball_xy_size_um / mean_in_plane_pixel_size))
    ball_z_size = int(round(ball_z_size_um / float(voxel_sizes[0])))

    return soma_diameter, max_cluster_size, ball_xy_size, ball_z_size


def main(
    signal_array,
    start_plane,
    end_plane,
    voxel_sizes,
    soma_diameter,
    max_cluster_size,
    ball_xy_size,
    ball_z_size,
    ball_overlap_fraction,
    soma_spread_factor,
    n_free_cpus,
    log_sigma_size,
    n_sds_above_mean_thresh,
    outlier_keep=False,
    artifact_keep=False,
    save_planes=False,
    plane_directory=None,
):
    n_processes = get_num_processes(min_free_cpu_cores=n_free_cpus)
    start_time = datetime.now()

    (
        soma_diameter,
        max_cluster_size,
        ball_xy_size,
        ball_z_size,
    ) = calculate_parameters_in_pixels(
        voxel_sizes,
        soma_diameter,
        max_cluster_size,
        ball_xy_size,
        ball_z_size,
    )

    if end_plane == -1:
        end_plane = len(signal_array)
    signal_array = signal_array[start_plane:end_plane]

    workers_queue = MultiprocessingQueue(maxsize=n_processes)
    # WARNING: needs to be AT LEAST ball_z_size
    mp_3d_filter_queue = MultiprocessingQueue(maxsize=ball_z_size)
    for plane_id in range(n_processes):
        # place holder for the queue to have the right size on first run
        workers_queue.put(None)

    setup_params = [
        signal_array[0, :, :],
        soma_diameter,
        ball_xy_size,
        ball_z_size,
        ball_overlap_fraction,
        start_plane,
    ]
    output_queue = MultiprocessingQueue()

    mp_3d_filter = Mp3DFilter(
        mp_3d_filter_queue,
        output_queue,
        soma_diameter,
        setup_params=setup_params,
        soma_size_spread_factor=soma_spread_factor,
        planes_paths_range=signal_array,
        save_planes=save_planes,
        plane_directory=plane_directory,
        start_plane=start_plane,
        max_cluster_size=max_cluster_size,
        outlier_keep=outlier_keep,
        artifact_keep=artifact_keep,
    )

    # start 3D analysis (waits for planes in queue)
    bf_process = multiprocessing.Process(target=mp_3d_filter.process, args=())
    bf_process.start()  # needs to be started before the loop
    clipping_val, threshold_value = setup_tile_filtering(signal_array[0, :, :])
    mp_tile_processor = MpTileProcessor(workers_queue, mp_3d_filter_queue)
    prev_lock = Lock()
    processes = []

    # start 2D tile filter (output goes into queue for 3D analysis)
    for plane_id, plane in enumerate(signal_array):
        workers_queue.get()
        lock = Lock()
        lock.acquire()
        p = multiprocessing.Process(
            target=mp_tile_processor.process,
            args=(
                plane_id,
                np.array(plane),
                prev_lock,
                lock,
                clipping_val,
                threshold_value,
                soma_diameter,
                log_sigma_size,
                n_sds_above_mean_thresh,
            ),
        )
        prev_lock = lock
        processes.append(p)
        p.start()

    processes[-1].join()
    mp_3d_filter_queue.put((None, None, None))  # Signal the end
    cells = output_queue.get()
    bf_process.join()

    print(
        "Detection complete - all planes done in : {}".format(
            datetime.now() - start_time
        )
    )
    return cells



from pkg_resources import resource_filename


def source_config_cellfinder():
    return resource_filename("cellfinder_core", "config/cellfinder.conf")


def source_custom_config_cellfinder():
    return resource_filename(
        "cellfinder_core", "config/cellfinder.conf.custom"
    )



import math

from cellfinder_core.detect.filters.volume.ball_filter import (
    BallFilter,
)
from cellfinder_core.detect.filters.volume.structure_detection import (
    CellDetector,
)
from cellfinder_core.tools.tools import get_max_value


def setup(
    plane,
    soma_diameter,
    ball_xy_size,
    ball_z_size,
    ball_overlap_fraction=0.6,
    z_offset=0,
):
    plane = plane.T

    max_value = get_max_value(plane)
    thrsh_val = max_value - 1
    soma_centre_val = max_value

    tile_width = soma_diameter * 2
    layer_width, layer_height = plane.shape

    ball_filter = BallFilter(
        layer_width,
        layer_height,
        ball_xy_size,
        ball_z_size,
        overlap_fraction=ball_overlap_fraction,
        tile_step_width=tile_width,
        tile_step_height=tile_width,
        threshold_value=thrsh_val,
        soma_centre_value=soma_centre_val,
    )
    start_z = z_offset + int(math.floor(ball_z_size / 2))
    cell_detector = CellDetector(layer_width, layer_height, start_z=start_z)

    return ball_filter, cell_detector


def setup_tile_filtering(plane):

    max_value = get_max_value(plane)
    clipping_value = max_value - 2
    thrsh_val = max_value - 1

    return clipping_value, thrsh_val



cdef extern from "Python.h":

    ###########################################################################
    # Codec registry and support functions
    ###########################################################################

    int PyCodec_Register(object search_function)
    # Register a new codec search function.

    # As side effect, this tries to load the encodings package, if not yet
    # done, to make sure that it is always first in the list of search
    # functions.

    int PyCodec_KnownEncoding(const char *encoding)
    # Return 1 or 0 depending on whether there is a registered codec for the
    # given encoding. This function always succeeds.

    object PyCodec_Encode(object o, const char *encoding, const char *errors)
    # Return value: New reference.
    # Generic codec based encoding API.

    # o is passed through the encoder function found for the given encoding
    # using the error handling method defined by errors. errors may be NULL
    # to use the default method defined for the codec. Raises a LookupError
    # if no encoder can be found.

    object PyCodec_Decode(object o, const char *encoding, const char *errors)
    # Return value: New reference.
    # Generic codec based decoding API.

    # o is passed through the decoder function found for the given encoding
    # using the error handling method defined by errors. errors may be NULL
    # to use the default method defined for the codec. Raises a LookupError
    # if no encoder can be found.


    # Codec lookup API

    # In the following functions, the encoding string is looked up converted
    # to all lower-case characters, which makes encodings looked up through
    # this mechanism effectively case-insensitive. If no codec is found, a
    # KeyError is set and NULL returned.

    object PyCodec_Encoder(const char *encoding)
    # Return value: New reference.
    # Get an encoder function for the given encoding.

    object PyCodec_Decoder(const char *encoding)
    # Return value: New reference.
    # Get a decoder function for the given encoding.

    object PyCodec_IncrementalEncoder(const char *encoding, const char *errors)
    # Return value: New reference.
    # Get an IncrementalEncoder object for the given encoding.

    object PyCodec_IncrementalDecoder(const char *encoding, const char *errors)
    # Return value: New reference.
    # Get an IncrementalDecoder object for the given encoding.

    object PyCodec_StreamReader(const char *encoding, object stream, const char *errors)
    # Return value: New reference.
    # Get a StreamReader factory function for the given encoding.

    object PyCodec_StreamWriter(const char *encoding, object stream, const char *errors)
    # Return value: New reference.
    # Get a StreamWriter factory function for the given encoding.


    # Registry API for Unicode encoding error handlers

    int PyCodec_RegisterError(const char *name, object error) except? -1
    # Register the error handling callback function error under the given
    # name. This callback function will be called by a codec when it
    # encounters unencodable characters/undecodable bytes and name is
    # specified as the error parameter in the call to the encode/decode
    # function.

    # The callback gets a single argument, an instance of
    # UnicodeEncodeError, UnicodeDecodeError or UnicodeTranslateError that
    # holds information about the problematic sequence of characters or bytes
    # and their offset in the original string (see Unicode Exception Objects
    # for functions to extract this information). The callback must either
    # raise the given exception, or return a two-item tuple containing the
    # replacement for the problematic sequence, and an integer giving the
    # offset in the original string at which encoding/decoding should be
    # resumed.

    # Return 0 on success, -1 on error.

    object PyCodec_LookupError(const char *name)
    # Return value: New reference.
    # Lookup the error handling callback function registered under name. As a
    # special case NULL can be passed, in which case the error handling
    # callback for "strict" will be returned.

    object PyCodec_StrictErrors(object exc)
    # Return value: Always NULL.
    # Raise exc as an exception.

    object PyCodec_IgnoreErrors(object exc)
    # Return value: New reference.
    # Ignore the unicode error, skipping the faulty input.

    object PyCodec_ReplaceErrors(object exc)
    # Return value: New reference.
    # Replace the unicode encode error with "?" or "U+FFFD".

    object PyCodec_XMLCharRefReplaceErrors(object exc)
    # Return value: New reference.
    # Replace the unicode encode error with XML character references.

    object PyCodec_BackslashReplaceErrors(object exc)
    # Return value: New reference.
    # Replace the unicode encode error with backslash escapes ("\x", "\u"
    # and "\U").

    object PyCodec_NameReplaceErrors(object exc)
    # Return value: New reference.
    # Replace the unicode encode error with "\N{...}" escapes.

    # New in version 3.5.



"""
N.B imports are within functions to prevent tensorflow being imported before
it's warnings are silenced
"""

import os
import logging
from imlib.general.logging import suppress_specific_logs

tf_suppress_log_messages = [
    "multiprocessing can interact badly with TensorFlow"
]


def main(
    signal_array,
    background_array,
    voxel_sizes,
    start_plane=0,
    end_plane=-1,
    trained_model=None,
    model_weights=None,
    model="resnet50_tv",
    batch_size=32,
    n_free_cpus=2,
    network_voxel_sizes=[5, 1, 1],
    soma_diameter=16,
    ball_xy_size=6,
    ball_z_size=15,
    ball_overlap_fraction=0.6,
    log_sigma_size=0.2,
    n_sds_above_mean_thresh=10,
    soma_spread_factor=1.4,
    max_cluster_size=100000,
    cube_width=50,
    cube_height=50,
    cube_depth=20,
    network_depth="50",
):
    suppress_tf_logging(tf_suppress_log_messages)

    from cellfinder_core.detect import detect
    from cellfinder_core.classify import classify
    from cellfinder_core.tools import prep
    from pathlib import Path

    home = Path.home()
    install_path = home / ".cellfinder"
    logging.info("Detecting cell candidates")

    points = detect.main(
        signal_array,
        start_plane,
        end_plane,
        voxel_sizes,
        soma_diameter,
        max_cluster_size,
        ball_xy_size,
        ball_z_size,
        ball_overlap_fraction,
        soma_spread_factor,
        n_free_cpus,
        log_sigma_size,
        n_sds_above_mean_thresh,
    )

    model_weights = prep.prep_classification(
        trained_model, model_weights, install_path, model, n_free_cpus
    )
    if len(points) > 0:
        logging.info("Running classification")
        points = classify.main(
            points,
            signal_array,
            background_array,
            n_free_cpus,
            voxel_sizes,
            network_voxel_sizes,
            batch_size,
            cube_height,
            cube_width,
            cube_depth,
            trained_model,
            model_weights,
            network_depth,
        )
    else:
        logging.info("No candidates, skipping classification")
    return points
    # logging.info("Saving classified cells")
    # save_cells(points, classified_points_path)


def suppress_tf_logging(tf_suppress_log_messages):
    """
    Prevents many lines of logs such as:
    "2019-10-24 16:54:41.363978: I tensorflow/stream_executor/platform/default
    /dso_loader.cc:44] Successfully opened dynamic library libcuda.so.1"
    """
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

    for message in tf_suppress_log_messages:
        suppress_specific_logs("tensorflow", message)



# distutils: language = c++

from libcpp.map cimport map as CppMap
from libcpp.vector cimport vector as CppVector
from libcpp.pair cimport pair as CppPair

from cellfinder_core.detect.filters.typedefs cimport ull, ulong, uint, ushort


cdef get_biggest_structure(CppVector[ulong] sizes):
    cdef uint result = 0
    cdef uint val
    for val in sizes:
        if val > result:
            result = val
    return result


cdef class BaseTileFilter:

    cdef:
        readonly uint out_of_brain_intensity_threshold
        public ushort[:,:] _tile
        public ushort current_threshold
        public bint keep
        SizeAnalyser size_analyser

    def __init__(self, out_of_brain_intensity_threshold=100):
        """

        :param int out_of_brain_intensity_threshold: Set to 0 to disable
        """
        self.out_of_brain_intensity_threshold = out_of_brain_intensity_threshold
        self.current_threshold = -1
        self.keep = True
        self.size_analyser = SizeAnalyser()

    cpdef set_tile(self, tile):
        raise NotImplementedError

    cpdef get_tile(self):
        raise NotImplementedError

    cpdef get_structures(self):
        cdef CppVector[ulong] struct_sizes
        self.size_analyser.process(self._tile, self.current_threshold)
        struct_sizes = self.size_analyser.get_sizes()
        return get_biggest_structure(struct_sizes), struct_sizes.size()

    cpdef is_low_average(self):  # TODO: move to OutOfBrainTileFilter
        cdef bint is_low
        cdef double avg = 0
        cdef uint x, y
        for x in range(self._tile.shape[0]):
            for y in range(self._tile.shape[1]):
                avg += self._tile[x, y]
        avg /= self._tile.shape[0] * self._tile.shape[1]
        is_low = avg < self.out_of_brain_intensity_threshold
        self.keep = not is_low
        return is_low


cdef class OutOfBrainTileFilter(BaseTileFilter):

    cpdef set_tile(self, tile):
        self._tile = tile

    cpdef get_tile(self):
        return self._tile


cdef class SizeAnalyser:

    cdef:
        CppMap[ull, ull] obsolete_ids
        CppMap[ull, ulong] struct_sizes

    cpdef process(self, ushort[:,:] tile, ushort threshold):
        tile = tile.copy()
        self.clear_maps()

        cdef ull struct_id, id_west, id_north, last_structure_id
        last_structure_id = 1

        cdef uint y, x
        for y in range(tile.shape[1]):
            for x in range(tile.shape[0]):
                # default struct_id to 0 so that it is not counted as structure in next iterations
                id_west = id_north = struct_id = 0
                if tile[x, y] >= threshold:
                    # If in bounds look to neighbours
                    if x > 0:
                        id_west = tile[x-1, y]
                    if y > 0:
                        id_north = tile[x, y-1]

                    id_west = self.sanitise_id(id_west)
                    id_north = self.sanitise_id(id_north)

                    if id_west != 0:
                        if id_north != 0 and id_north != id_west:
                            struct_id = self.merge_structures(id_west, id_north)
                        else:
                            struct_id = id_west
                    elif id_north != 0:
                        struct_id = id_north
                    else:  # no neighbours, create new structure
                        struct_id = last_structure_id
                        self.struct_sizes[last_structure_id] = 0
                        last_structure_id += 1

                    self.struct_sizes[struct_id] += 1

                tile[x, y] = struct_id

    cpdef get_sizes(self):
        cdef CppVector[ulong] struct_sizes
        cdef ulong size
        cdef ull s_id
        cdef CppPair[ull, ulong] iterator_pair
        for iterator_pair in self.struct_sizes:
            struct_sizes.push_back(iterator_pair.second)
        return struct_sizes

    cdef clear_maps(self):
        self.obsolete_ids.clear()
        self.struct_sizes.clear()

    cdef sanitise_id(self, ull s_id):
        while self.obsolete_ids.count(s_id):  # walk up the chain of obsolescence
            s_id = self.obsolete_ids[s_id]
        return s_id

    # id1 and id2 must be valid struct IDs (>0)!
    cdef merge_structures(self, ull id1, ull id2):
        cdef ulong new_size
        cdef ull tmp_id

        # ensure id1 is the smaller of the two values
        if id2 < id1:
            tmp_id = id1    # swap
            id1 = id2
            id2 = tmp_id

        self.struct_sizes[id1] += self.struct_sizes[id2]
        self.struct_sizes.erase(id2)

        self.obsolete_ids[id2] = id1

        return id1
